import yaml
import sys
import os
from agno.tools.function import Function

from colored import cprint
from a1facts.ontology.property import Property
from a1facts.ontology.entity_class import EntityClass
from a1facts.ontology.relationship_class import RelationshipClass
from a1facts.ontology.ontology_rewrite_agent import OntologyRewriteAgent
from a1facts.utils.telemetry import nonblocking_send_telemetry_ping
from a1facts.utils.logger import logger
import re
from a1facts.ontology.rdfs_entity import RDFSEntity
from a1facts.ontology.rdfs_relationship import RDFSRelationship
from a1facts.ontology.rdfs_property import RDFSProperty
from rdflib import Graph, Literal, BNode, URIRef
from rdflib.namespace import RDF

class KnowledgeOntology:
    """
    Represents the entire ontology, including all entity and relationship classes.
    It loads the ontology from a YAML file and provides methods to access its components.
    """
    def __init__(self, ontology_file: str):
        logger.system(f"Initializing KnowledgeOntology for {ontology_file}")
        """
        Initializes the KnowledgeOntology object.

        Args:
            ontology_file (str): The path to the YAML file defining the ontology.
        """
        self.ontology_file = ontology_file
        self.entity_classes = []
        self.relationship_classes = []
        self.name = ""
        self.description = ""
        self.main_entities = []
        logger.system(f"Loading ontology from {ontology_file}")
        self.load_ontology()
        logger.system(f"Ontology loaded from {ontology_file}")
        self.rewrite_agent = OntologyRewriteAgent(self.ontology_file, [])
        logger.system(f"Ontology rewrite agent initialized")
        nonblocking_send_telemetry_ping()
        logger.user(f"Ontology loaded: {self.ontology_file}")
        logger.user(f"Ontology parsed: {str(self)}")
        
    def find_relationship_class(self, relationship_class_name):
        """
        Finds a relationship class by name.

        Args:
            relationship_class_name (str): The name of the relationship class to find.
        """
        logger.system(f"Finding relationship class: {relationship_class_name}")
        for relationship_class in self.relationship_classes:
            if relationship_class.relationship_name == relationship_class_name:
                return relationship_class
        logger.system(f"Relationship class not found: {relationship_class_name}")
        return None
        
    def find_entity_class(self, entity_class_name):
        """
        Finds an entity class by name.

        Args:
            entity_class_name (str): The name of the entity class to find.

        Returns:
            EntityClass or None: The found entity class, or None if not found.
        """
        logger.system(f"Finding entity class: {entity_class_name}")
        for entity_class in self.entity_classes:
            if entity_class.entity_class_name == entity_class_name:
                return entity_class
        logger.system(f"Entity class not found: {entity_class_name}")
        return None   
 


    def load_ontology(self):
        """Loads the ontology from the specified YAML file."""
        logger.system(f"Loading ontology from {self.ontology_file}")
        with open(self.ontology_file, 'r') as file:
            ontology = yaml.load(file, Loader=yaml.FullLoader)
            world_data = ontology.get('world', {})
            self.name = world_data.get('name', 'N/A')
            self.description = world_data.get('description', 'N/A')
            self.main_entities = world_data.get('main_entities')
            if self.main_entities is None:
                raise ValueError(f"'main_entities' is a required field in the 'world' section of the ontology file: {self.ontology_file}")

            for name, details in ontology.get('entity_classes', {}).items():
                entity_class = EntityClass(name, details.get('description', 'N/A'))
                entity_class.properties = []                
                for prop in details.get('properties', []):
                    entity_class.add_property(Property(prop.get('name', 'N/A'), prop.get('type', 'N/A'), prop.get('description', 'N/A'), prop.get('primary_key', False)))
                self.entity_classes.append(entity_class)
            for name, details in ontology.get('relationships', {}).items():
                domain = self.find_entity_class(details.get('domain', 'N/A'))
                range = self.find_entity_class(details.get('range', 'N/A'))
                symmetric = details.get('symmetric', False)
                relationship_class = RelationshipClass(name, domain, range, details.get('description', 'N/A'), symmetric)
                self.relationship_classes.append(relationship_class)
        logger.system(f"Ontology loaded from {self.ontology_file}")

    def get_tools_add_or_update_entity(self, add_entity_func):
        """
        Gets a list of all add/update tools for all entity classes.

        Args:
            add_entity_func (function): The function to call to add/update an entity.

        Returns:
            list: A list of tool functions.
        """
        logger.system(f"Getting entity add/update tools")
        tools = []
        for entity_class in self.entity_classes:
            tools.append(entity_class.get_tool_add_or_update_entity(add_entity_func))
        logger.system(f"Entity add/update tools returned")
        return tools

    def get_tools_get_entity_properties(self, get_entity_properties_func):
        """
        Gets a list of all 'get properties' tools for all entity classes.

        Args:
            get_entity_properties_func (function): The function to call to get entity properties.

        Returns:
            list: A list of tool functions.
        """
        logger.system(f"Getting entity get properties tools")
        tools = []
        for entity_class in self.entity_classes:
            tools.append(entity_class.get_tool_get_entity_properties(get_entity_properties_func))
        logger.system(f"Entity get properties tools returned")
        return tools

    def get_tools_get_all_entity(self, get_all_entity_func):
        """
        Gets a list of all 'get all' tools for all entity classes.

        Args:
            get_all_entity_func (function): The function to call to get all entities.

        Returns:
            list: A list of tool functions.
        """
        logger.system(f"Getting entity get all tools")
        tools = []
        for entity_class in self.entity_classes:
            tools.append(entity_class.get_tool_get_all_entity(get_all_entity_func))
        logger.system(f"Entity get all tools returned")
        return tools

    def get_tools_add_or_update_relationship(self, add_relationship_func):
        """
        Gets a list of all add/update tools for all relationship classes.

        Args:
            add_relationship_func (function): The function to call to add/update a relationship.

        Returns:
            list: A list of tool functions.
        """
        logger.system(f"Getting relationship add/update tools")
        tools = []
        for relationship_class in self.relationship_classes:
            tools.append(relationship_class.get_tool_add_or_update_relationship(add_relationship_func))
        logger.system(f"Relationship add/update tools returned")
        return tools

    def get_tools_get_relationship_entities(self, get_relationship_entities_func):
        """
        Gets a list of all 'get related entities' tools for all relationship classes.

        Args:
            get_relationship_entities_func (function): The function to call to get related entities.

        Returns:
            list: A list of tool functions.
        """
        logger.system(f"Getting relationship get relationship entities tools")
        tools = []
        for relationship_class in self.relationship_classes:
            tools.append(relationship_class.get_tool_get_relationship_entities(get_relationship_entities_func))
        logger.system(f"Relationship get relationship entities tools returned")
        return tools

    def get_tools_add_or_update_entity_and_relationship(self, add_entity_func, add_relationship_func):
        """
        Gets a combined list of all add/update tools for both entities and relationships.

        Args:
            add_entity_func (function): The function to call to add/update an entity.
            add_relationship_func (function): The function to call to add/update a relationship.

        Returns:
            list: A list of all add/update tool functions.
        """
        logger.system(f"Getting add/update tools")
        tools = []
        tools.extend(self.get_tools_add_or_update_entity(add_entity_func))
        tools.extend(self.get_tools_add_or_update_relationship(add_relationship_func))
        logger.system(f"Add/update tools returned")
        return tools

    def get_tools_get_entity_and_relationship(self, get_all_entity_func, get_entity_properties_func, get_relationship_entities_func):
        """
        Gets a combined list of all 'get' tools for both entities and relationships.

        Args:
            get_all_entity_func (function): Function to get all entities.
            get_entity_properties_func (function): Function to get entity properties.
            get_relationship_entities_func (function): Function to get related entities.

        Returns:
            list: A list of all 'get' tool functions.
        """
        logger.system(f"Getting get tools")
        tools = []
        tools.extend(self.get_tools_get_all_entity(get_all_entity_func))
        tools.extend(self.get_tools_get_entity_properties(get_entity_properties_func))
        tools.extend(self.get_tools_get_relationship_entities(get_relationship_entities_func))
        logger.system(f"Get tools returned")
        return tools

    def __str__(self):
        """Returns a string representation of the entire ontology."""
        logger.system(f"Getting string representation of ontology")
        ontology_str = ""
        ontology_str += f"Ontology Name: {self.name}\n"
        ontology_str += f"Ontology Description: {self.description}\n"
        ontology_str += "Entity Classes:\n"
        for entity_class in self.entity_classes:
            ontology_str += f"   {entity_class}\n"
        ontology_str += "Relationship Classes:\n"
        for relationship_class in self.relationship_classes:
            ontology_str += f"   {relationship_class}"
        return ontology_str

    def parse_rdfs_with_validation(self, rdfs_content: str):
        """
        Parses and validates an RDFS string against the ontology using rdflib.
        
        Args:
            rdfs_content: A string containing the RDFS data in Turtle format.
        
        Returns:
            A tuple containing two lists: (entities, relationships)
        """
        entities = {}
        relationships = []
        error_messages = []

        # Find the base URI from prefixes, if it exists
        base_uri_match = re.search(r'@prefix\s*:\s*<([^>]+)>', rdfs_content)
        base_uri = base_uri_match.group(1) if base_uri_match else None

        processed_rdfs = rdfs_content.strip()
        if processed_rdfs.startswith('{') and processed_rdfs.endswith('}'):
            processed_rdfs = processed_rdfs[1:-1].strip()

        # Ensure essential prefixes are present
        if '@prefix rdfs:' not in processed_rdfs:
            processed_rdfs = '@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .\n' + processed_rdfs
        if '@prefix xsd:' not in processed_rdfs:
            processed_rdfs = '@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .\n' + processed_rdfs
        if base_uri and '@prefix :' not in processed_rdfs:
            processed_rdfs = f'@prefix : <{base_uri}> .\n' + processed_rdfs

        g = Graph()
        
        try:
            g.parse(data=processed_rdfs, format='turtle')
        except Exception as e:
            error_messages.append(f"RDFS parsing error: {e}")
            logger.system("\n--- ❌ Validation Errors ---")
            for msg in error_messages:
                logger.system(msg)
            return [], []


        def get_alias(term, graph):
            if isinstance(term, URIRef):
                return graph.qname(term)
            return str(term)

        # 1. First pass: identify and create all entities with their properties
        for s, o in g.subject_objects(predicate=RDF.type):
            entity_name = get_alias(s, g)
            entity_class_name = get_alias(o, g)
            entity_class = self.find_entity_class(entity_class_name)
            if not entity_class:
                error_messages.append(f"VALIDATION ERROR: Entity class '{entity_class_name}' not found in ontology for entity '{entity_name}'.")
                continue

            rdfs_properties = []
            for p_prop, o_prop in g.predicate_objects(subject=s):
                if isinstance(o_prop, Literal):  # This is a property
                    prop_key = get_alias(p_prop, g)
                    prop_value = str(o_prop)
                    rdfs_properties.append(RDFSProperty(prop_key, prop_value))
            if entity_class.validate_properties(rdfs_properties, entity_name, error_messages, 0):
                properties_dict = {prop.key: prop.value for prop in rdfs_properties}
                properties_dict['type'] = entity_class_name
                entity = RDFSEntity(entity_class=entity_class, entity_name=entity_name, properties=properties_dict)
                entities[entity_name] = entity
        # 2. Second pass: identify and create relationships
        for s, p, o in g:
            # Skip entity declarations and properties of entities
            if p == RDF.type or isinstance(o, Literal):
                continue

            relationship_name = get_alias(p, g)
            relationship_class = self.find_relationship_class(relationship_name)

            if relationship_class:
                domain_name = get_alias(s, g)
                domain_entity = entities.get(domain_name)
                
                range_entity = None
                rdfs_properties = []

                # This is a direct relationship without properties
                range_name = get_alias(o, g)
                range_entity = entities.get(range_name)


                if not domain_entity:
                    if domain_name not in entities:
                        error_messages.append(f"VALIDATION ERROR: Domain entity '{domain_name}' for relationship '{relationship_name}' not defined or failed validation.")
                    continue
                
                if not range_entity:
                    error_messages.append(f"VALIDATION ERROR: Range for relationship '{relationship_name}' must be an entity. Found '{range_name}'.")
                    continue

                valid = True
                if not relationship_class.validate_domain_and_range(domain_name, range_name, entities, self, error_messages, 0):
                    valid = False
                
                if valid:
                    symmetric = relationship_class.symmetric
                    rel = RDFSRelationship(domain_entity=domain_entity, relationship=relationship_name, range_entity=range_entity, symmetric=symmetric)
                    relationships.append(rel)

        # 3. Log results
        success_messages = [f"Created entity: {entity}" for entity in entities.values()]
        for rel in relationships:
            success_messages.append(f"Created relationship: {rel}")

        logger.system("\n--- ✅ Successful Operations ---")
        for msg in success_messages:
            logger.system(msg)

        if error_messages:
            logger.system("\n--- ❌ Validation Errors ---")
            for msg in error_messages:
                logger.system(msg)
        logger.system("\n--- Parsing and Validation Complete ---")

        return list(entities.values()), relationships



from a1facts.ontology.knowledge_ontology import KnowledgeOntology
from a1facts.utils.modelconfig import my_query_model, my_fast_tool_calling_model
from agno.agent import Agent
from textwrap import dedent
from datetime import datetime
from colored import cprint
from a1facts.utils.logger import logger

class QueryAgent:
    def __init__(self, ontology: KnowledgeOntology, mytools: list):
        self.ontology = ontology        
        self.agent = Agent(
            name="Knowledge Graph Agent",
            role="Interact with the knowledge graph",
            model=my_fast_tool_calling_model,
            tools=mytools,
            instructions=dedent(f"""
                Get information from the knowledge base.
                Use the tools to get information from the graph.
                Today is {datetime.now().strftime("%Y-%m-%d")}

                Only use information from the knowledge graph to answer the question, do not use your own knowledge, do not make up answers. 
                If you don't know the answer, say "A verifiable answer is not available" - don't add any other text.

                Provide all sources for your answer, the sources should be extracted from the properties of the entities in the knowledge graph; you should get them when you get the information from the graph.          
                """),
                markdown=True,
                debug_mode=False,
            )

    def query(self, query: str):
        # Try to answer with the knowledge graph first
        result = self.agent.run(dedent(f"""Answer the question in %QUERY%, 
        Use as many tools as you need to get the information necessary to answer the query.
        %QUERY%: {query}
        Today is {datetime.now().strftime("%Y-%m-%d")}
        """))     

        if not result.content:
            logger.system(f"No answer found to {query}, returning fallback.")
            return "A verifiable answer is not available."
        
        return result.content


from __future__ import annotations
from a1facts.ontology.entity_class import EntityClass
from a1facts.ontology.property import Property
import typing
if typing.TYPE_CHECKING:
    from a1facts.ontology.knowledge_ontology import KnowledgeOntology


class RDFSEntity:
    """Represents an RDFS entity."""

    def __init__(self, entity_class: EntityClass, entity_name: str, properties: dict):
        """
        Initializes an RDFSEntity object.

        Args:
            entity_class (EntityClass): The entity class of the entity.
            entity_name (str): The name of the entity.
            properties (dict): A dictionary of properties for the entity.
        """
        self.entity_class = entity_class
        self.name = entity_name
        self.properties = properties

    def __str__(self):
        """Returns a string representation of the RDFS entity."""
        props_str = ", ".join(f"{k}: {v}" for k, v in self.properties.items())
        return f"RDFSEntity(name={self.name}, properties={{{props_str}}})"

    def print(self):
        """Prints the string representation of the RDFS entity."""
        print(str(self))



from a1facts.graph.knowledge_graph import KnowledgeGraph
from a1facts.ontology.knowledge_ontology import KnowledgeOntology
from a1facts.enrichment.knowledge_acquirer import KnowledgeAcquirer
from colored import cprint
from a1facts.utils.logger import logger
from a1facts.utils.timer import timer

class KnowledgeBase:
    def __init__(self, name: str, ontology_config_file: str, knowledge_sources_config_file: str, use_neo4j: bool = False, disable_exa: bool = False, graph_file: str = "networkx_graph.pickle", neo4j_uri: str = None, neo4j_user: str = None, neo4j_password: str = None):
        logger.system(f"Initializing KnowledgeBase for {name}")
        self.name = name
        self.ontology = KnowledgeOntology(ontology_config_file)
        self.graph = KnowledgeGraph(
            self.ontology,
            use_neo4j,
            graph_file=graph_file,
            neo4j_uri=neo4j_uri,
            neo4j_user=neo4j_user,
            neo4j_password=neo4j_password
        )
        self.knowledge_acquirer = KnowledgeAcquirer(self.graph, self.ontology, knowledge_sources_config_file, disable_exa)
        logger.system(f"KnowledgeBase initialized for {self.name}")

    def query(self, query: str):
        """
        Executes a query against the knowledge graph.

        This method allows you to retrieve information stored in the knowledge graph
        based on its ontological structure.

        Args:
            query (str): The query to execute against the knowledge graph.

        Returns:
            str: The result of the query from the knowledge graph.
        """
        logger.user(f"Querying knowledge graph for {query}")
        cprint(f"Querying knowledge graph", "green")
        truncated_query = query[:70] + "..." if len(query) > 70 else query
        cprint(f"Query: {truncated_query}", "yellow")
        return self.graph.query(query)

    def acquire_knowledge_for_query(self, query: str):
        """
        Acquires new knowledge based on a query and updates the knowledge graph.

        This method uses the configured knowledge sources to find new information
        related to the query, which is then integrated into the knowledge graph.

        Args:
            query (str): The query to guide the knowledge acquisition process.

        Returns:
            str: The newly acquired knowledge.
        """
        logger.user(f"Acquiring knowledge for {query}")
        cprint(f"Acquiring knowledge", "green")
        truncated_query = query[:70] + "..." if len(query) > 70 else query
        cprint(f"Knowledge seeked: {truncated_query}", "yellow")
        newknowledge = self.knowledge_acquirer.acquire(query)
        self.ingest_knowledge(newknowledge)
        return newknowledge
    
    def ingest_knowledge(self, knowledge: str):
        """
        Ingests and integrates new knowledge into the knowledge graph.

        This method takes a string of knowledge, processes it, and updates the
        knowledge graph according to the ontology.

        Args:
            knowledge (str): The knowledge to be ingested into the knowledge graph.

        Returns:
            str: The result of the knowledge ingestion operation.
        """
        logger.user(f"Ingesting knowledge for {knowledge}")
        cprint(f"Ingesting knowledge", "green")
        truncated_knowledge = knowledge[:70] + "..." if len(knowledge) > 70 else knowledge
        cprint(f"Knowledge to update: {truncated_knowledge}", "yellow")
        return self.graph.update_knowledge(knowledge)

    def get_tools(self):
        logger.system(f"Getting tools for {self.name}")
        def query_tool(query: str):
            return self.query(query)

        def acquire_tool(query: str):
            result = self.acquire_knowledge_for_query(query)
            return result

        query_tool.__doc__ = f"""Query the knowledge graph for precise information for {self.ontology.description}
ARGS: query: str - The query to query the knowledge graph
RETURNS: str - The result from the knowledge graph"""
        query_tool.__parameters__ = {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "The query to query the knowledge graph"
                }
            },
            "required": ["query"]
        }

        acquire_tool.__doc__ = f"""Acquire knowledge from the knowledge acquirer and update the knowledge graph for {self.ontology.description}
Objective: To get high reliability and credibility information from the knowledge sources
ARGS: query: str - The query to acquire knowledge from the knowledge acquirer
RETURNS: str - The result from the knowledge acquirer"""
        acquire_tool.__parameters__ = {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "The query to acquire knowledge from the knowledge acquirer"
                }
            },
            "required": ["query"]
        }
        logger.system(f"Tools returned for {self.name}")
        return [query_tool, acquire_tool]

    def __str__(self) -> str:
        return f"a1facts('{self.name}', ontology='{self.ontology}', knowledge_acquirer={self.knowledge_acquirer})"

    def __del__(self) -> None:
        """
        Destructor that automatically closes the Neo4j graph connection
        when the a1facts instance is garbage collected.
        """
        logger.system(f"Destroying KnowledgeBase for {self.name}")
        if hasattr(self, 'graph') and self.graph:
            self.graph.close()
            logger.system(f"KnowledgeBase closed for {self.name}")
        timer.print_results()


import spacy
from spacy.language import Language
from spacy.tokens import Span
from sentence_transformers import SentenceTransformer
import chromadb
import yaml
import uuid

# --- 1. Dynamic, Learning Knowledge Base ---
class DynamicKnowledgeBase:
    """
    Manages a knowledge base that is built dynamically from processed text.
    It encapsulates the "find or create" logic for entity resolution.
    """
    def __init__(self, ontology_yaml: str, model_name='all-MiniLM-L6-v2', similarity_threshold=0.85):
        print("Initializing Dynamic Knowledge Base...")
        self.ontology = yaml.safe_load(ontology_yaml)
        self.main_entities = self.ontology.get('world', {}).get('main_entities', [])
        self.entity_counter = {entity: 0 for entity in self.main_entities}
        
        # In-memory storage for the dynamic KB
        self.alias_to_id = {}
        self.id_to_canonical = {}
        
        # Setup for semantic search
        self.model = SentenceTransformer(model_name)
        self.similarity_threshold = similarity_threshold
        chroma_client = chromadb.Client()
        self.collection = chroma_client.get_or_create_collection(
            name="dynamic_entity_kb",
            metadata={"hnsw:space": "cosine"}
        )
        print(f"Ready to resolve main entities: {self.main_entities}")

    def _create_new_entity(self, text: str, entity_type: str) -> str:
        """Adds a new entity to the knowledge base."""
        # Determine the entity type from NER label (e.g., ORG -> Company)
        # This mapping might need to be more sophisticated in a real system.
        # For now, we'll assume a direct mapping if possible.
        kb_type = entity_type if entity_type in self.main_entities else "Unknown"
        
        self.entity_counter[kb_type] = self.entity_counter.get(kb_type, 0) + 1
        new_id = f"{kb_type.upper()}_{self.entity_counter[kb_type]}"
        
        print(f"  -> CREATING new entity. ID: {new_id}, Canonical Name: '{text}'")
        
        # Update KB state
        self.id_to_canonical[new_id] = text
        self.alias_to_id[text.lower()] = new_id
        
        # Index the new entity in ChromaDB for future semantic lookups
        embedding = self.model.encode([text]).tolist()
        self.collection.add(
            embeddings=embedding,
            documents=[text],
            metadatas=[{"canonical_id": new_id}],
            ids=[str(uuid.uuid4())]
        )
        return new_id

    def find_or_create_entity(self, text: str, entity_type: str) -> (str, str):
        """
        The core logic: finds an existing entity or creates a new one.
        Returns the canonical ID and name.
        """
        text_lower = text.lower().strip()
        
        # 1. Exact Match
        if text_lower in self.alias_to_id:
            kb_id = self.alias_to_id[text_lower]
            print(f"  -> FOUND entity via Exact Match. ID: {kb_id}")
            return kb_id, self.id_to_canonical[kb_id], "Exact"
            
        # 2. Semantic Match
        if self.collection.count() > 0:
            query_embedding = self.model.encode([text]).tolist()
            results = self.collection.query(query_embeddings=query_embedding, n_results=1)
            
            if results and results['distances'][0]:
                distance = results['distances'][0][0]
                similarity = 1 - distance
                if similarity >= self.similarity_threshold:
                    kb_id = results['metadatas'][0][0]['canonical_id']
                    print(f"  -> FOUND entity via Semantic Match. ID: {kb_id} (Score: {similarity:.2f})")
                    # Learn this new alias!
                    print(f"     Learning new alias: '{text}' -> {self.id_to_canonical[kb_id]}")
                    self.alias_to_id[text_lower] = kb_id
                    return kb_id, self.id_to_canonical[kb_id], f"Semantic (Score: {similarity:.2f})"

        # 3. Create New Entity
        kb_id = self._create_new_entity(text, entity_type)
        return kb_id, text, "Created New"


# --- 2. spaCy Component (now simpler) ---
@Language.factory(
    "dynamic_entity_linker",
    default_config={
        "ontology_yaml": "",
        "model_name": "all-mpnet-base-v2",
        "similarity_threshold": 0.85,
    },
)
def create_dynamic_entity_linker(
    nlp: Language,
    name: str,
    ontology_yaml: str,
    model_name: str,
    similarity_threshold: float,
):
    """Factory function to create the DynamicEntityLinker component."""
    kb = DynamicKnowledgeBase(
        ontology_yaml=ontology_yaml,
        model_name=model_name,
        similarity_threshold=similarity_threshold,
    )
    return DynamicEntityLinker(kb=kb)


class DynamicEntityLinker:
    def __init__(self, kb: DynamicKnowledgeBase):
        self.kb = kb
        for ext in ["kb_id", "canonical_name", "match_method"]:
            if not Span.has_extension(ext):
                Span.set_extension(ext, default=None)

    def __call__(self, doc):
        for ent in doc.ents:
            # Simple mapping from spaCy labels to ontology types.
            # This can be customized for more complex scenarios.
            entity_type_map = {"ORG": "Company", "PRODUCT": "Drug", "PERSON": "Sponsor"}
            ent_type = entity_type_map.get(ent.label_, ent.label_)

            # Only resolve entities defined as 'main' in the ontology
            if ent_type in self.kb.main_entities:
                kb_id, canonical_name, method = self.kb.find_or_create_entity(ent.text, ent_type)
                ent._.kb_id = kb_id
                ent._.canonical_name = canonical_name
                ent._.match_method = method
        return doc


# --- 3. Main Execution Block ---
if __name__ == "__main__":
    financial_ontology_yaml = """
    world:
      name: Company Knowledge Graph
      main_entities: [Company]
    entity_classes:
      Company:
        properties: [{name: name, type: str, primary_key: true}]
    """
    
    # 1. Setup spaCy pipeline
    nlp = spacy.load("en_core_web_lg")
    nlp.add_pipe("dynamic_entity_linker", after="ner", config={
        "ontology_yaml": financial_ontology_yaml,
        "similarity_threshold": 0.85,
    })

    # 2. Process texts SEQUENTIALLY to see the KB learn
    texts_to_process = [
        "A report mentioned Intel Corporation as a key player.",
        "Later, Intel Inc. released its earnings.",
        "Separately, UnitedHealth Group announced a new initiative.",
        "A statement from the healthcare company United Health followed."
    ]

    print("\n" + "="*50)
    print("Processing documents and building Knowledge Base dynamically...")
    print("="*50)

    for i, text in enumerate(texts_to_process):
        print(f"\n--- Processing Doc {i+1}: '{text}' ---")
        doc = nlp(text)
        for ent in doc.ents:
            if ent._.kb_id: # Check if it was resolved by our component
                print(
                    f"  Entity: '{ent.text}' (Label: {ent.label_})\n"
                    f"  -> Resolved ID: {ent._.kb_id}\n"
                    f"  -> Canonical Name: {ent._.canonical_name}\n"
                    f"  -> Method: {ent._.match_method}"
                )
    
    # Get the component from the pipeline to access its knowledge base
    dynamic_linker = nlp.get_pipe("dynamic_entity_linker")
    kb = dynamic_linker.kb

    print("\n" + "="*50)
    print("Final Knowledge Base State:")
    print("="*50)
    print("Canonical Entities:", kb.id_to_canonical)
    print("Learned Aliases:", kb.alias_to_id)


from a1facts.graph.knowledge_graph import KnowledgeGraph
from a1facts.ontology.knowledge_ontology import KnowledgeOntology
from a1facts.enrichment.knowledge_source import KnowledgeSource
from a1facts.enrichment.function_knowledge_source.knowledge_source import FunctionKnowledgeSource
from a1facts.enrichment.mcp_knowledge_source.knowledge_source import MCPKnowledgeSource
from a1facts.utils.modelconfig import my_high_precision_model
from agno.agent import Agent
from textwrap import dedent
from agno.tools.exa import ExaTools
from datetime import datetime
import yaml
from colored import cprint
from a1facts.utils.logger import logger
import pickle
import os
import hashlib


class KnowledgeAcquirer:
    def __init__(self, graph: KnowledgeGraph, ontology: KnowledgeOntology, knowledge_sources_config_file: str, disable_exa: bool = False):
        logger.user(f"Initializing Knowledge Sources for {knowledge_sources_config_file} with disable_exa: {disable_exa}")
        self.ontology = ontology
        self.graph = graph
        self.knowledge_sources = self.load_knowledge_sources(knowledge_sources_config_file)
        logger.system(f"Knowledge sources loaded")
        for source in self.knowledge_sources:
            logger.system(f"Knowledge source loaded: {source.name}")
        self.tools = []
        if not disable_exa:
            self.tools.append(ExaTools(num_results=20, summary=True))
            logger.system(f"Exa tools loaded")
        for source in self.knowledge_sources:
            self.tools.append(source.query_tool())
            logger.system(f"Knowledge source query tool loaded: {source.name}")
        self.tools.append(self.graph.get_tools)
        self.agent = Agent(
            name="Knowledge Acquirer",
            role=dedent("""Enrich and update the knowledge graph with validated information from the knowledge sources.
            Always provide the sources for the answer. Never make up sources.
            If the sources come from the knowledge graph, provide the actual content of the "sources" property of the entity/relationship.
            Never cite the knowledge graph itself as a source.
            """),
            model=my_high_precision_model,
            tools=self.tools,
            instructions=self.get_acquisition_instructions(),
            markdown=True,
            debug_mode=False,
        )
        logger.user(f"KnowledgeAcquirer initialized")
        cprint(f"KnowledgeAcquirer initialized", "green")

    def get_acquisition_instructions(self):

        cache_file = 'acquirer_instructions.pickle'
        ontology_str = str(self.ontology)
        current_ontology_hash = hashlib.sha256(ontology_str.encode('utf-8')).hexdigest()
        
        if os.path.exists(cache_file):
            try:
                with open(cache_file, 'rb') as f:
                    cached_data = pickle.load(f)

                if cached_data.get('ontology_hash') == current_ontology_hash:
                    logger.system(f"Using cached acquisition instructions")
                    return cached_data['instructions']
            except (pickle.UnpicklingError, EOFError, KeyError) as e:
                # Handle cases where the pickle file is corrupt or has unexpected format
                print(f"Cache file {cache_file} is invalid, regenerating. Error: {e}")
        # Cache miss or invalid cache file
        instructions = self.ontology.rewrite_agent.rewrite_query(self.get_template())
        with open(cache_file, 'wb') as f:
            pickle.dump({
                'ontology_hash': current_ontology_hash,
                'instructions': instructions
            }, f)
        logger.system(f"Acquisition instructions cached")

        return instructions

    def acquire(self, query: str):
        result = self.agent.run(query)
        cprint(f"KnowledgeAcquirer result: {result.content}", "green")
        return result.content

    def load_knowledge_sources(self, knowledge_sources_config_file: str):
        knowledge_sources = []
        logger.system(f"Loading knowledge sources from {knowledge_sources_config_file}")
        with open(knowledge_sources_config_file, 'r') as file:
            knowledge_sources_config = yaml.load(file, Loader=yaml.FullLoader)
            logger.system(f"Knowledge sources config loading from {knowledge_sources_config}")
            if knowledge_sources_config.get('knowledge_sources'):
                for source in knowledge_sources_config['knowledge_sources']:
                    if 'type' not in knowledge_sources_config['knowledge_sources'][source]:
                        logger.warning(f"Your knowledge source config is missing the 'type' field for source: {source}")
                        raise ValueError(f"Your knowledge source config is missing the 'type' field for source: {source}")
                    source_type = knowledge_sources_config['knowledge_sources'][source]['type']
                    if source_type == 'function':
                        source_config = knowledge_sources_config['knowledge_sources'][source]
                        source = FunctionKnowledgeSource(source_config)
                        knowledge_sources.append(source)
                    elif source_type == 'mcp':
                        source_config = knowledge_sources_config['knowledge_sources'][source]
                        source = MCPKnowledgeSource(source_config)
                        knowledge_sources.append(source)
                    else: 
                        logger.warning(f"Unknown knowledge source type: {source_type}")
            else: 
                logger.warning(f"No knowledge sources found in {knowledge_sources_config_file}")
        return knowledge_sources

    def get_template(self):
        return dedent(f"""(Template Instructions: Before use, replace the bracketed placeholders [...] with the specific details relevant to your target ontology and knowledge base.)

Goal:
Expand the knowledge graph to answer the user's query with validated and reliable information.
Return the full answer to the user's query.

Procedure:
1. Query the knowledge graph to retrieve relevant entities and relationships.    
1.2. If the knowledge graph lacks the necessary information(entities, properties, relationships or outdated information), query the knowledge sources (including the web) to supplement the knowledge graph.
1.3. If the knowledge graph has the necessary information, and the information is guaranteed to be current, return the answer to the user's query.
1.4. If the knowledge graph has the necessary information, but the information is not guaranteed to be current, query the knowledge sources (including the web) to supplement the knowledge graph.
1.5. If the knowledge graph lacks the necessary information, and the information is not guaranteed to be current, query the knowledge sources (including the web) to supplement the knowledge graph.
1.6. If the knowledge sources lack the necessary information, return "A verifiable answer is not available".

2. Data Supplementation & Vetting
Prioritization: Always prioritize foundational data that pertains to [Main entity classes] and [Main relationship types] from the [knowledge graph]

Web Search Scope: Use web searches to find:
1. Real-time or very recent data not available in the knowledge base.
2. The latest news, press releases, or official filings.
3. Broader contextual information (e.g., economic indicators, industry trends).
4. Third-party analysis (e.g., analyst ratings, reviews).

For information coming from knowledge sources that have a reliability rating of A or B, you can use the information directly.

IMPORTANT:
3. Source Reliability Assessment: Evaluate each web source using the following scale: [Keep these verbatim in your transformation to the ontology given]
    A: Completely reliable - The source is undoubtedly authentic and trustworthy.
    B: Usually reliable - Minor doubts exist, but the source is historically valid.
    C: Fairly reliable - Doubts exist, but the source has provided valid information before.
    D: Not usually reliable - Significant doubts about the source's reliability.
    E: Unreliable - The source has a history of providing invalid information.
    F: Reliability cannot be judged - Insufficient information for evaluation.
Constraint: Never use sources rated D, E, or F.

IMPORTANT:
4. Information Synthesis & Validation [Keep these verbatim in your transformation to the ontology given]
Information Validity Assessment: After gathering information, assess each piece of data across all sources using this scale:
    1. Confirmed: Corroborated by multiple, independent, reliable sources.
    2. Probably true: Logical and consistent with other data, but not fully corroborated.
    3. Possibly true: Plausible but lacks strong corroboration.
    4. Doubtful: Not logical or may be contradicted by other information.
    5. Improbable: Illogical and contradicted by other information.
    6. Cannot be judged: Insufficient information to assess validity.
Constraint: Only use information assessed as Confirmed, Probably true, or Possibly true.

Timeliness: Ensure all data is current relative to the date below.

5. Output Formatting
Data Display: Use tables to present quantitative and comparative data.
Use bullet points to present qualitative and descriptive data.

Citation: When information is Confirmed by multiple sources, cite those sources in your response.
Always provide sources for your answer, the sources should be extracted from the properties of the entities in the knowledge graph; you should get them when you get the information from the graph.          
Current Date: {datetime.now().strftime("%Y-%m-%d")}
""")



from __future__ import annotations
from datetime import datetime

def check_type(value: str, expected_type: str) -> bool:
    """Validates if a string value can be cast to the expected type."""
    if expected_type == 'string' or expected_type == 'str':
        return True
    if expected_type == 'float':
        try:
            float(value)
            return True
        except ValueError:
            return False
    if expected_type == 'integer' or expected_type == 'int':
        try:
            int(value)
            return True
        except ValueError:
            return False
    if expected_type == 'bool':
        return value.lower() in ['true', 'false']
    if expected_type == 'date':
        try:
            datetime.fromisoformat(value)
            return True
        except (ValueError, TypeError):
            return False
    return False

class RDFSProperty:
    def __init__(self, key: str, value: str):
        self.key = key
        self.value = value



from a1facts.graph.graph_database import BaseGraphDatabase
from neo4j import GraphDatabase
from a1facts.utils.logger import logger
from dotenv import load_dotenv
from a1facts.ontology.rdfs_entity import RDFSEntity
from a1facts.ontology.rdfs_relationship import RDFSRelationship

import os
from colored import cprint
from datetime import date


load_dotenv()
URI = os.getenv("NEO4J_URI")
AUTH = ("neo4j", os.getenv("NEO4J_AUTH"))

class Neo4jGraphDatabase(BaseGraphDatabase):
    def __init__(self, uri=None, user=None, password=None):
        try:
            db_uri = uri or URI
            if user and password:
                db_auth = (user, password)
            else:
                db_auth = (AUTH[0], AUTH[1])

            self.driver = GraphDatabase.driver(db_uri, auth=db_auth)
            cprint("Successfully connected to Neo4j database.", "green")
        except Exception as e:
            print(f"Failed to connect to Neo4j database: {e}")
            self.driver = None
    def _execute_query(self, query, parameters=None):
        """
        Executes a Cypher query that writes data to the graph.

        Args:
            query (str): The Cypher query to execute.
            parameters (dict, optional): Parameters for the query. Defaults to None.
        """
        if self.driver is None:
            print("Driver not initialized. Cannot execute query.")
            return

        with self.driver.session() as session:
            try:
                session.run(query, parameters)
            except Exception as e:
                print(f"Error executing query: {e}")

    def _execute_read_query(self, query, parameters=None):
        """
        Executes a Cypher query that reads data from the graph.

        Args:
            query (str): The Cypher query to execute.
            parameters (dict, optional): Parameters for the query. Defaults to None.
        
        Returns:
            list: A list of records from the query result.
        """
        if self.driver is None:
            print("Driver not initialized. Cannot execute query.")
            return []

        with self.driver.session() as session:
            try:
                result = session.run(query, parameters)
                return [record for record in result]
            except Exception as e:
                print(f"Error executing read query: {e}")
                return []

    def close(self):
        if self.driver is not None:
            self.driver.close()

    def add_or_update_entity(self, entity: RDFSEntity):
        label = entity.entity_class.entity_class_name
        primary_key_field = entity.entity_class.primary_key_prop.property_name
        properties = entity.properties
        
        if primary_key_field not in properties:
            print(f"Error: Primary key '{primary_key_field}' not found in properties.")
            return

        sanitized_props = {}
        for key, value in properties.items():
            if isinstance(value, date):
                sanitized_props[key] = value.isoformat()
            else:
                sanitized_props[key] = value

        primary_value = sanitized_props[primary_key_field]

        # Cypher query to find a node by its label and primary key, or create it if it doesn't exist.
        # ON CREATE sets all properties when the node is first created.
        # ON MATCH updates all properties if the node already exists.
        query = (
            f"MERGE (n:{label} {{{primary_key_field}: $primary_value}}) "
            "ON CREATE SET n = $props "
            "ON MATCH SET n += $props"
        )

        parameters = {
            "primary_value": primary_value,
            "props": sanitized_props
        }

        self._execute_query(query, parameters)
        #print(f"Successfully added/updated entity: {label} with {primary_key_field} = '{primary_value}'")

    def add_relationship(self, relationship: RDFSRelationship):
        start_node_label = relationship.domain_entity.entity_class.entity_class_name
        start_pk_field = relationship.domain_entity.entity_class.primary_key_prop.property_name
        start_node_pk_val = relationship.domain_entity.properties[start_pk_field]
        end_node_label = relationship.range_entity.entity_class.entity_class_name
        end_pk_field = relationship.range_entity.entity_class.primary_key_prop.property_name
        end_node_pk_val = relationship.range_entity.properties[end_pk_field]
        relationship_type = relationship.relationship
        symmetric = relationship.symmetric

        # Base query for a directional relationship
        query = (
            f"MATCH (a:{start_node_label} {{{start_pk_field}: $start_val}}), "
            f"(b:{end_node_label} {{{end_pk_field}: $end_val}}) "
            f"MERGE (a)-[r:{relationship_type}]->(b) "
        )

        # If the relationship is symmetric, create the reverse relationship as well
        if symmetric:
            reverse_query = (
                f"MATCH (a:{start_node_label} {{{start_pk_field}: $start_val}}), "
                f"(b:{end_node_label} {{{end_pk_field}: $end_val}}) "
                f"MERGE (b)-[r:{relationship_type}]->(a) "
            )
        
        parameters = {
            "start_val": start_node_pk_val,
            "end_val": end_node_pk_val
        }

        try:
            self._execute_query(query, parameters)
            if symmetric:
                self._execute_query(reverse_query, parameters)
        except Exception as e:
            print(f"Error creating relationship: {e}")
            return False
        
        #print(f"Successfully created relationship: ({start_node_pk_val})-[{relationship_type}]->({end_node_pk_val})")

    def _get_primary_key_field(self, label):
        """
        Determines the primary key field for a given entity label.

        Args:
            label (str): The label of the entity.

        Returns:
            str: The name of the primary key field.
        """
        if label == "Role":
            return "role_title"
        return "name"

    def find_entities_fuzzy(self, label, search_field, search_term):
        """
        Finds entities using a case-insensitive, partial string match.

        Args:
            label (str): The label of the node to search for.
            search_field (str): The property field to search within (e.g., "name").
            search_term (str): The term to search for.

        Returns:
            list: A list of nodes that match the search criteria.
        """
        query = (
            f"MATCH (n:{label}) "
            f"WHERE toLower(n.{search_field}) CONTAINS toLower($search_term) "
            "RETURN n"
        )
        parameters = {"search_term": search_term}
        
        results = self._execute_read_query(query, parameters)
        
        # Extract the node data from the result records
        if not results:
            return []
        nodes = [record["n"] for record in results]
        return nodes

    def get_entity_info(self, label, entity_identifier, exact_match=False):
        """
        Retrieves properties of an entity and the names of entities it's related to.

        Args:
            label (str): The label for the node (e.g., "Organization").
            entity_identifier (str): The name or title of the entity to query.
            exact_match (bool): If True, performs an exact match. Defaults to False (fuzzy).

        Returns:
            list: A list of dictionaries, each containing an entity's properties and relationships.
        """
        pk_field = self._get_primary_key_field(label)

        if exact_match:
            where_clause = f"n.{pk_field} = $identifier"
        else:
            where_clause = f"toLower(n.{pk_field}) CONTAINS toLower($identifier)"

        query = (
            f"MATCH (n:{label}) "
            f"WHERE {where_clause} "
            "OPTIONAL MATCH (n)-[r]-(related) "
            "RETURN properties(n) AS properties, "
            "collect({relationship: type(r), related_entity: coalesce(related.name, related.role_title)}) AS relationships"
        )
        parameters = {"identifier": entity_identifier}
        records = self._execute_read_query(query, parameters)

        if not records:
            print(f"No entity with label '{label}' and identifier '{entity_identifier}' found.")
            return []

        results = []
        for record in records:
            if record["properties"]:
                # Filter out potential empty relationship objects if no relationships exist
                relationships = [rel for rel in record["relationships"] if rel['relationship'] is not None]
                results.append({"properties": record["properties"], "relationships": relationships})
        
        return results

    def get_all_entities_by_label(self, label):
        """
        Retrieves all entities (nodes) with a specific label from the graph.

        Args:
            label (str): The label to search for (e.g., "Organization").

        Returns:
            list: A list of dictionaries, where each represents an entity's properties.
        """
        query = f"MATCH (n:{label}) RETURN properties(n) AS properties"
        records = self._execute_read_query(query)

        if not records:
            print(f"No entities found with label '{label}'.")
            return []
        
        return [record["properties"] for record in records]

    def get_relationship_entities(self, domain_label, domain_pk_prop, domain_primary_key_value, relationship_type, range_label):
        """
        Gets all range entities connected to a specific domain entity via a relationship.

        Args:
            domain_label (str): The label of the domain entity.
            domain_pk_prop (str): The primary key property of the domain entity.
            domain_primary_key_value (str): The primary key of the domain entity.
            relationship_type (str): The type of the relationship.
            range_label (str): The label of the range entities to retrieve.

        Returns:
            list: A list of dictionaries, where each represents a range entity's properties.
        """
        # For a given domain, get all the range entities in a relationship
        query = f"MATCH (n:{domain_label} {{{domain_pk_prop}: $domain_primary_key_value}}) MATCH (n)-[r:{relationship_type}]->(m:{range_label}) RETURN properties(m) AS properties"
        parameters = {"domain_primary_key_value": domain_primary_key_value}
        records = self._execute_read_query(query, parameters)
        return [record["properties"] for record in records]
    
    def get_entity_properties(self, label, pk_prop, primary_key_value):
        """
        Gets the properties of a single entity identified by its primary key.

        Args:
            label (str): The label of the entity.
            pk_prop (str): The primary key property of the entity.
            primary_key_value (str): The primary key value of the entity.

        Returns:
            dict or None: The properties of the entity, or None if not found.
        """
        # For a given entity, get the properties
        query = f"MATCH (n:{label} {{{pk_prop}: $primary_key_value}}) RETURN properties(n) AS properties"
        parameters = {"primary_key_value": primary_key_value}
        records = self._execute_read_query(query, parameters)
        if records:
            return records[0]["properties"]
        return None



import time
from colored import cprint

class Timer:
    def __init__(self):
        self.records = {}

    def start(self, label):
        if label not in self.records:
            self.records[label] = {}
        self.records[label]['start'] = time.time()

    def stop(self, label):
        if label not in self.records or 'start' not in self.records[label]:
            raise ValueError(f"Timer for '{label}' was not started.")
        self.records[label]['duration'] = time.time() - self.records[label]['start']

    def print_results(self):
        for label, record in self.records.items():
            if 'duration' in record:
                cprint(f"{label}: {record['duration']:.4f}s", "red")

timer = Timer()



from __future__ import annotations
from a1facts.ontology.rdfs_entity import RDFSEntity
import typing
if typing.TYPE_CHECKING:
    from a1facts.ontology.knowledge_ontology import KnowledgeOntology


class RDFSRelationship:
    """Represents an RDFS relationship."""

    def __init__(self, domain_entity: RDFSEntity, relationship: str, range_entity: RDFSEntity, symmetric: bool = False):
        """
        Initializes an RDFSRelationship object.

        Args:
            domain_entity (RDFSEntity): The domain of the relationship.
            relationship (str): The relationship itself.
            range_entity (RDFSEntity): The range of the relationship.
        """
        self.domain_entity = domain_entity
        self.relationship = relationship
        self.range_entity = range_entity
        self.symmetric = symmetric
        
    def __str__(self):
        """Returns a string representation of the RDFS relationship."""
        return f"RDFSRelationship(domain={self.domain_entity.name}, relationship={self.relationship}, range={self.range_entity.name})"

    def print(self):
        """Prints the string representation of the RDFS relationship."""
        print(str(self))



from a1facts.ontology.knowledge_ontology import KnowledgeOntology
from a1facts.utils.modelconfig import my_high_precision_model
from agno.agent import Agent
from textwrap import dedent
from datetime import datetime
from a1facts.utils.logger import logger
from colored import cprint
from pydantic import BaseModel, Field


class RDFSResult(BaseModel):
    rdfs: str = Field(description="The RDFS format of the knowledge. Only include entities with their properties and relationships that are in the ontology.")
    other_information: str = Field(description="Other information that couldn't be described in the ontology. Ideally this is empty, but if you have other info that is not in the ontology, included it here verbatim.")
    ontology_elements_used: list[str] = Field(description="The elements of the ontology that were used to create the RDFS format. each line should represent either an entity (with properties) or a relationship used from the ontology.")


class UpdateAgent:
    def __init__(self, ontology: KnowledgeOntology, graph_database):
        self.ontology = ontology
        self.graph_database = graph_database
        self.rdfs_agent = Agent(
            name="RDFS Agent",
            role="Translate the knowledge into a structured format based on the ontology.",
            model=my_high_precision_model,
            instructions=dedent(f"""
                Translate the knowledge into a structured format based on the ontology.
                Ontology:[{self.ontology}]
                Return the results in turtle RDFS format. Include entities with their properties and relationships.
                ALWAYS use the properties verbatim that were used from the ontology.
                ALWAYS use the entities and relationships verbatim that were used from the ontology.
                NEVER make up information. 
                Only use the information provided to you. DO NOT use your own knowledge to make up information.
                Ideally, every RDFS entity should have sources.
                If you have an entity that can't be expressed in the ontology, include it in the other_information field.
                Use ":" as prefix for the entities and relationships. use "a" to describe an entity that is a type of another entity.
                Today is {datetime.now().strftime("%Y-%m-%d")}
                The following is an example of the RDFS format:
                @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
                @prefix : <http://example.org/ontology#> .
                @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

                # --- Define the entities ---

                :Mars a :Planet ;
                    :commonName "Mars" .

                # --- Define the Observation (the relationship node) and its properties ---

                :Opposition_Event_2025 a :AstronomicalObservation ;
                    :observationID "Opposition_Event_2025" ;
                    :observationType "Opposition" ;
                    :magnitude "-2.8"^^xsd:decimal ;
                    # These properties belong to the event, NOT Mars
                    :observationDate "2025-12-08"^^xsd:date ;
                    :visibility "Excellent" .

                # --- Link the entities together ---

                # This single triple connects Mars to its detailed observation event.
                :Mars :hasObservation :Opposition_Event_2025 .
            """),
            markdown=True,
            debug_mode=False,
            output_schema=RDFSResult,
        )
        

    def update(self, knowledge: str) -> str:
        #step 0 - translate the knowledge into rdfs format
        #step 1 - deduplicate the knowledge using spacy; add alises to entities and relatioinships
        #step 2 - add the knowledge to the graph using the tools available to you
        logger.system(f"Updating knowledge graph with knowledge: {knowledge}")
        rdfs_result = self.rdfs_agent.run("Translate the following knowledge into a structured format based on the ontology\n\n " + knowledge)
        
        if not rdfs_result or not rdfs_result.content or not rdfs_result.content.rdfs:
            logger.system("No RDFS content was generated by the agent.")
            return "No information was updated."
            
        logger.system("\n--- RDFS Content ---")
        logger.system(rdfs_result.content.rdfs)

        logger.system("\n--- Other Information ---") #TODO: this can be used to improve the ontology
        logger.system(rdfs_result.content.other_information)      
        try:
            entities, relationships = self.ontology.parse_rdfs_with_validation(rdfs_result.content.rdfs)
        except Exception as e:
            logger.system(f"An error occurred during RDFS parsing and validation: {e}")
            return f"Failed to update knowledge due to a parsing error: {e}"
        logger.system("\n--- Parsed Entities ---")
        return_str = ""
        for entity in entities:
            self.graph_database.add_or_update_entity(entity)
            logger.system(entity)
            return_str += f"Added entity: {entity}\n"
        logger.system("\n--- Parsed Relationships ---")
        for relationship in relationships:
            self.graph_database.add_relationship(relationship)
            logger.system(relationship)
            return_str += f"Added relationship: {relationship}\n"

        return return_str


from a1facts.ontology.knowledge_ontology import KnowledgeOntology
from a1facts.utils.modelconfig import my_fast_language_model
from agno.agent import Agent
from textwrap import dedent
from datetime import datetime
from colored import cprint
from a1facts.utils.logger import logger




class QueryRewriteAgent:
    def __init__(self, ontology: KnowledgeOntology, mytools: list):
        self.ontology = ontology        
        self.agent = Agent(
            name="Query rewrite agent",
            role="Rewrite the query to use known entities",
            model=my_fast_language_model,
            tools=mytools,
            instructions=dedent(f"""
                Rewrite the query to use known entities.
                Today is {datetime.now().strftime("%Y-%m-%d")}
                """),
                markdown=True,
                debug_mode=False,
            )
    
    def rewrite_query(self, query: str, class_entity_pairs: dict):

        logger.system(f"Rewriting query: {query}")
        prompt = dedent(f"""
            Rewrite the query to use known entities from the graph, for entities not in the graph keep the entity names as is.
            Here are the known entity pairs: {class_entity_pairs}
            Here is the query to rewrite: {query}
            Preserve the original meaning of the query, and the original form of the query.
            Only return the rewritten query, no other text.
            """
        )

        result = self.agent.run(prompt)
        result_content = result.content
        result_content = result_content+f"\nToday is {datetime.now().strftime("%Y-%m-%d")}"
        cprint("Rewritten query: " + result_content, 'green')
        logger.system(f"Rewritten query: {result_content}")
        return result_content


from a1facts.graph.graph_database import BaseGraphDatabase
import networkx as nx
import pickle
from colored import cprint
from a1facts.utils.logger import logger
from io import open
from a1facts.ontology.rdfs_entity import RDFSEntity
from a1facts.ontology.rdfs_relationship import RDFSRelationship


class NetworkxGraphDatabase(BaseGraphDatabase):
    def __init__(self, graph_file="networkx_graph.pickle"):
        self.graph = nx.DiGraph()
        self.graph_file = graph_file
        try:
            with open(self.graph_file, "rb") as f:
                self.graph = pickle.load(f)
        except FileNotFoundError:
            pass
        num_nodes = self.graph.number_of_nodes()
        num_edges = self.graph.number_of_edges()
        cprint(f"Successfully initialized Networkx database with {num_nodes} nodes, {num_edges} relationships.", "green")
        self.nodes_by_label = self._build_label_index()

    def _build_label_index(self):
        index = {}
        for node, data in self.graph.nodes(data=True):
            label = data.get('label')
            if label:
                if label not in index:
                    index[label] = set()
                index[label].add(node)
        return index

    def add_or_update_entity(self, entity: RDFSEntity):
        label = entity.entity_class.entity_class_name
        primary_key_field = entity.entity_class.primary_key_prop.property_name
        properties = entity.properties
        logger.system(f"NWX: Adding or updating {label} entity with primary key {primary_key_field} and properties {properties}")
        if primary_key_field not in properties:
            logger.system(f"NWX: Primary key '{primary_key_field}' not found in properties.")
            return

        primary_key_value = properties[primary_key_field]
        # Use a tuple of (label, pk_value) as the unique node identifier
        node_id = (label, primary_key_value)
        
        node_properties = properties.copy()
        node_properties['label'] = label

        if self.graph.has_node(node_id):
            self.graph.nodes[node_id].update(node_properties)
        else:
            self.graph.add_node(node_id, **node_properties)
            
        # Add to new label set in index
        if label not in self.nodes_by_label:
            self.nodes_by_label[label] = set()
        self.nodes_by_label[label].add(node_id)


    def add_relationship(self, relationship: RDFSRelationship):
        start_node_label = relationship.domain_entity.entity_class.entity_class_name
        start_pk_field = relationship.domain_entity.entity_class.primary_key_prop.property_name
        start_node_pk_val = relationship.domain_entity.properties[start_pk_field]
        end_node_label = relationship.range_entity.entity_class.entity_class_name
        end_pk_field = relationship.range_entity.entity_class.primary_key_prop.property_name
        end_node_pk_val = relationship.range_entity.properties[end_pk_field]
        relationship_type = relationship.relationship

        symmetric = relationship.symmetric

        logger.system(f"NWX: Adding {relationship_type} relationship between {start_node_label} {start_node_pk_val} and {end_node_label} {end_node_pk_val}")
        
        start_node_id = (start_node_label, start_node_pk_val)
        end_node_id = (end_node_label, end_node_pk_val)
        
        edge_properties = {}
        edge_properties['type'] = relationship_type

        self.graph.add_edge(start_node_id, end_node_id, **edge_properties)
        if symmetric:
            self.graph.add_edge(end_node_id, start_node_id, **edge_properties)

    def get_all_entities_by_label(self, label):
        logger.system(f"NWX: Getting all {label} entities")
        # This logic is now simpler as we can iterate through all nodes
        # and check their label property directly.
        return [data for node, data in self.graph.nodes(data=True) if data.get('label') == label]

    def get_relationship_entities(self, domain_label, domain_pk_prop, domain_primary_key_value, relationship_type, range_label, range_primary_key_prop):
        logger.system(f"NWX: Getting {relationship_type} relationship entities for {domain_label} {domain_primary_key_value} and {range_label}")
        results = []
        domain_node_id = (domain_label, domain_primary_key_value)
        
        if not self.graph.has_node(domain_node_id):
            logger.system(f"NWX: No domain node found for {domain_label} {domain_primary_key_value}")
            return results

        for neighbor in self.graph.successors(domain_node_id):
            edge_data = self.graph.get_edge_data(domain_node_id, neighbor)
            if (edge_data and edge_data.get('type') == relationship_type and
                    self.graph.nodes[neighbor].get('label') == range_label):
                results.append(self.graph.nodes[neighbor])
        return results

    def get_entity_properties(self, label, pk_prop, primary_key_value):
        logger.system(f"NWX: Getting {label} properties for {primary_key_value}")
        node_id = (label, primary_key_value)
        if self.graph.has_node(node_id):
            return self.graph.nodes[node_id]
        else:
            logger.system(f"NWX: No node found for {label} {primary_key_value}")
            return None

    def print_graph(self):
        logger.system(f"NWX: Printing graph")
        print("All nodes in the graph:")
        for node, data in self.graph.nodes(data=True):
            print(f"  Node: {node}, Data: {data}")
        
        print("\nAll relationships in the graph:")
        for start, end, data in self.graph.edges(data=True):
            print(f"  Edge: {start} -> {end}, Data: {data}")

    def close(self):
        logger.system(f"NWX: Closing graph")
        self.graph = nx.DiGraph()
        self.nodes_by_label = {}
        #self.print_graph()

    def save(self):
        try:
            logger.system(f"NWX: Saving graph to {self.graph_file}")
            with open(self.graph_file, "wb") as f:
                pickle.dump(self.graph, f)
        except Exception as e:
            logger.system(f"Error saving graph to {self.graph_file}: {e}")
            print(f"Error saving graph to {self.graph_file}: {e}")